// Do not remove the include below

#include <Arduino.h>
#include "RaceMaster_Display_V0.h"

#include <ESP8266WiFi.h>
#include <pgmspace.h>
#include <Wire.h>  // must be incuded here so that Arduino library object file references work
#include <RtcDS1307.h>
#include <DHT.h>
#include <stdio.h>

#include <queue>

#define DHTPIN 13     // what pin we're connected to
#define DHTTYPE DHT11   // DHT 11
DHT dht(DHTPIN, DHTTYPE);

RtcDS1307 Rtc;

#define countof(a) (sizeof(a) / sizeof(a[0]))
#define MCP2301_MODULES 3
#define MCP_BUSS_ID_21 0x21
//#define SSID "racemaster"
#define SSID "wifiruss"
#define PASS "00112233445566778899aabbcc"

//Actual numbers
byte displayMap[] = {
	B00111111, B00000110, B01011011, B01001111, B01100110,
	B01101101, B01111101, B00000111, B01111111, B01100111,
	B10000000, B00000000
};

//Single pixels for testing
byte displayMapBits[] = { B00000000, B00000001, B00000010, B00000100, B00001000, B00010000,
B00100000, B01000000, B10000000 };

struct DisplaySegment{
	byte Segment;
	byte Port;
};
typedef struct {
	bool ready;
	double mhp;
	double et;
	double rt;
	bool showMPH;
	bool showET;
	bool showRT;
	bool showArrow;
} DisplayQueue;

DisplayQueue displayQ[2];

DisplaySegment segments[5];
void buildSegmentArray(){

	segments[0].Segment = 0x20;
	segments[0].Port = GPIOA;
	segments[1].Segment = 0x20;
	segments[1].Port = GPIOB;

	segments[2].Segment = 0x21;
	segments[2].Port = GPIOA;
	segments[3].Segment = 0x21;
	segments[3].Port = GPIOB;

	segments[4].Segment = 0x22;
	segments[4].Port = GPIOA;
	//Not used at this time but might be used for decimals?
	segments[5].Segment = 0x22;
	segments[5].Port = GPIOB;

}

IPAddress netServerIp = { 192, 168, 32, 200 };
IPAddress netMaskIp = { 255, 255, 255, 0 };
IPAddress netGWIp = { 192, 168, 32, 1 };
IPAddress netClientIp{ 192, 168, 50, 3 }; //1st Display's IP

void expanderWriteSingle(int port, const byte reg, const byte data)
{
	Wire.beginTransmission(port);
	Wire.write(reg); //register to write data to
	Wire.write(data);  // data to write
	Wire.endTransmission();
	delay(5);
} // end of expanderWrite
// set register "reg" on expander to "data" in byte mode
// for example, IO direction
void expanderWriteBoth(int port, const byte reg, const byte data)
{
	Wire.beginTransmission(port);
	Wire.write(reg);
	Wire.write(data);  // port A
	Wire.write(data);  // port B
	Wire.endTransmission();

	delay(5);
} // end of expanderWrite

void printDateTime(const RtcDateTime& dt)
{
	char datestring[20];
	snprintf(datestring, 20, "%02u/%02u/%04u %02u:%02u:%02u",
		dt.Month(),
		dt.Day(),
		dt.Year(),
		dt.Hour(),
		dt.Minute(),
		dt.Second());

	Serial.println(datestring);


}
void ShowDht(){

	// Reading temperature or humidity takes about 250 milliseconds!
	// Sensor readings may also be up to 2 seconds 'old' (its a very slow sensor)
	float h = dht.readHumidity();
	// Read temperature as Celsius (the default)
	float t = dht.readTemperature();
	// Read temperature as Fahrenheit (isFahrenheit = true)
	float f = dht.readTemperature(true);

	// Check if any reads failed and exit early (to try again).
	if (isnan(h) || isnan(t) || isnan(f)) {
		Serial.println("Failed to read from DHT sensor!");
		return;
	}

	// Compute heat index in Fahrenheit (the default)
	float hif = dht.computeHeatIndex(f, h);
	// Compute heat index in Celsius (isFahreheit = false)
	float hic = dht.computeHeatIndex(t, h, false);

	Serial.print("Humidity: ");
	Serial.print(h);
	Serial.print(" %\t");
	Serial.print("Temperature: ");
	Serial.print(t);
	Serial.print(" *C ");
	Serial.print(f);
	Serial.print(" *F\t");
	Serial.print("Heat index: ");
	Serial.print(hic);
	Serial.print(" *C ");
	Serial.print(hif);
	Serial.println(" *F");
}
void showRTC(){

	if (!Rtc.IsDateTimeValid())
	{
		// Common Cuases:
		//    1) the battery on the device is low or even missing and the power line was disconnected
		Serial.println("RTC lost confidence in the DateTime!");
	}

	RtcDateTime now = Rtc.GetDateTime();
	int d = now.Second();
	d = std::max(0, d % 10);
	//showDigit(MCP_BUSS_ID_21, d);

	printDateTime(now);

}

//set all segments in display to blank
void clearAllSegments(){
	for (int i = 0; i < 5; i++){
		expanderWriteSingle(segments[i].Segment, segments[i].Port, displayMap[11]);
	}
}
//Pad specified string
String stringPADL(String s, unsigned int finalLength){
	String t;
	for (int i = 0; i < (finalLength - s.length()); i++){
		t += " ";
	}
	return (t + s);


String stringPADR(String s, unsigned int finalLength){
	String t;
	for (int i = 0; i < (finalLength - s.length()); i++){
		t += " ";
	}
	return (s + t);
}
void setupRTC()
{
	//"Dec 26 2009", time = "12:34:56"
	//RtcDateTime compiled = RtcDateTime(__DATE__, __TIME__);
	RtcDateTime compiled = RtcDateTime("Jun 30 2015", "04:33:00");
	Rtc.SetDateTime(compiled);
	if (!Rtc.IsDateTimeValid())
	{
		// Common Cuases:
		//    1) first time you ran and the device wasn't running yet
		//    2) the battery on the device is low or even missing

		Serial.println("RTC lost confidence in the DateTime!");

		// following line sets the RTC to the date & time this sketch was compiled
		// it will also reset the valid flag internally unless the Rtc device is
		// having an issue

		Rtc.SetDateTime(compiled);
		RtcDateTime now = Rtc.GetDateTime();
		if (now < compiled)
		{
			Serial.println("RTC is older than compile time!  (Updating DateTime)");
			Rtc.SetDateTime(compiled);
		}
		else if (now > compiled)
		{
			Serial.println("RTC is newer than compile time. (this is expected)");
		}
		else if (now == compiled)
		{
			Serial.println("RTC is the same as compile time! (not expected but all is fine)");
		}
		Rtc.SetSquareWavePin(DS1307SquareWaveOut_Low);
	}
}


//Reverse bits pattern in specified byte
unsigned char reverse(unsigned char b) {
	b = (b & 0xF0) >> 4 | (b & 0x0F) << 4;
	b = (b & 0xCC) >> 2 | (b & 0x33) << 2;
	b = (b & 0xAA) >> 1 | (b & 0x55) << 1;
	return b;
}
void setupMCP(){
	//Setup all MCP chips for no sequential and all ports as output
	for (int i = 0; i < 3; i++){
		expanderWriteBoth(0x20 + i, IOCON, B00100000); //Turn off sequential mode
		expanderWriteSingle(0x20 + i, IODIRA, 0x00); //Set Direction to output for A
		expanderWriteSingle(0x20 + i, IODIRB, 0x00); //Set Direction to output for B
		expanderWriteSingle(0x20 + i, GPIOA, 0x00); //Clear PORTS A
		expanderWriteSingle(0x20 + i, GPIOB, 0x00); //Clear PORTS B
	}
}

WiFiServer server(21);

void setup()
{



	Serial.begin(115200);
	Wire.begin(5, 4);

	Rtc.Begin(); //RealTimeClock
	dht.begin(); //DHT11, Temp sensor

	setupRTC();
	setupMCP(); //MCP CHIPS
	buildSegmentArray();	//Build the display mapping structure array

	WiFi.disconnect();

	WiFi.stopSmartConfig(); //turn off smart config
	WiFi.disconnect(); //disconnect if smart config was previously on
	WiFi.mode(WIFI_STA); //Station mode
	WiFi.begin(SSID, PASS); //try to start
	//WiFi.config(netServerIp, netGWIp, netMaskIp);
	WiFi.config(netServerIp, netGWIp, netMaskIp);
	while (WiFi.status() != WL_CONNECTED){
		//Wait for client to connect
		//Flash the display for waiting network connection feedback
		expanderWriteSingle(0x20, GPIOA, B01000000);
		expanderWriteSingle(0x22, GPIOA, B01000000);
		delay(250);
		clearAllSegments();
		delay(250);
	}

	//	Serial.println(WiFi.localIP());
	//}

	server.begin(); //Start the tcp/ip server

}

unsigned long lastDigitSentTime = millis();
unsigned long lastShowTime = millis();

WiFiClient wc;



//Show elapsed time
// MPH: xxx.xx
//  ET: xx.xxx
//  RT: xx.xxx
void showET_OR_RT(double val){

	Serial.println(val,3);
	//Clear all the segments
	clearAllSegments();
	//Always pad left for a max of 5 digits
	//Note the point is not included in the digits
	//Pad string left to 5 characters
	String s = String(val, 3); //second parameter is number of digits after the decimal
	//Remove the decimal from the string
	s.replace(".", ""); //we know it is at the 1000's position (3rd from the right)

	//Now pad left for max of 5 characters
	// PAD so numbers can be right aligned on the display board
	// Simple skip spaces and use the appropriate segments with the "for indexer"
	s = stringPADL(s, 5); //Returns string padded right 5
	int sL = s.length();

	for (int i = 0; i < sL; i++)
	{
		//Don't show spaces/blanks
		if (s[i] != ' '){
			if (segments[i].Port == GPIOA){
				//Dont need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, displayMap[(int)(s[i] - 48)]);
			}
			else{//B port has pins wired backwards so have to reverse the bits
				//Need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, reverse(displayMap[(int)(s[i] - 48)]));
			}
		}
	}
	//Now need to show the right decimal
	//Second panel port B
	expanderWriteSingle(segments[1].Segment, segments[1].Port, reverse(displayMapBits[8] | displayMap[(int)(s[1] - 48)]));
}
//Show elapsed time
// MPH: xxx.xx
//  ET: xx.xxx
//  RT: xx.xxx
void showMPH(double val){
	Serial.print("MPH: ");
	Serial.println(val);
	//Clear all the segments
	clearAllSegments();
	//Always pad left for a max of 5 digits
	//Note the point is not included in the digits
	//Pad string left to 5 characters
	String s = String(val, 2); //second parameter is number of digits after the decimal
	//Remove the decimal from the string
	s.replace(".", ""); //we know it is at the 1000's position (3rd from the right)

	//Now pad left for max of 5 characters
	// PAD so numbers can be right aligned on the display board
	// Simple skip spaces and use the appropriate segments with the "for indexer"
	s = stringPADL(s, 5); //Returns string padded right 5
	int sL = s.length();

	for (int i = 0; i < sL; i++)
	{
		//Don't show spaces/blanks
		if (s[i] != ' '){
			if (segments[i].Port == GPIOA){
				//Dont need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, displayMap[(int)(s[i] - 48)]);
			}
			else{//B port has pins wired backwards so have to reverse the bits
				//Need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, reverse(displayMap[(int)(s[i] - 48)]));
			}
		}
	}
	//Now need to show the right decimal
	//Second panel port B, This segment don't need to be reversed
	expanderWriteSingle(segments[2].Segment, segments[2].Port, displayMapBits[7] | displayMap[(int)(s[1] - 48)]);
}
//Show the specified arrow indicating winner or looser
void showArrow(uint8_t whichArrow){


	Serial.println(whichArrow);

	clearAllSegments();
	if (whichArrow == 0){ //Left Lane
		expanderWriteSingle(segments[0].Segment, segments[0].Port, displayMapBits[7] | displayMap[10]);
		//Lines
		//expanderWriteSingle(segments[0].Segment, segments[0].Port, displayMapBits[10]);
		expanderWriteSingle(segments[1].Segment, segments[1].Port, reverse(displayMapBits[7]));
		expanderWriteSingle(segments[2].Segment, segments[2].Port, displayMapBits[7]);
		expanderWriteSingle(segments[3].Segment, segments[3].Port, reverse(displayMapBits[7]));
		expanderWriteSingle(segments[4].Segment, segments[4].Port, displayMapBits[7]);

	}
	else{ //Right Lane
		expanderWriteSingle(segments[4].Segment, segments[4].Port, displayMapBits[7] | displayMap[10]);
		//Lines
		expanderWriteSingle(segments[0].Segment, segments[0].Port, displayMapBits[7]);
		expanderWriteSingle(segments[1].Segment, segments[1].Port, reverse(displayMapBits[7]));
		expanderWriteSingle(segments[2].Segment, segments[2].Port, displayMapBits[7]);
		expanderWriteSingle(segments[3].Segment, segments[3].Port, reverse(displayMapBits[7]));
		//expanderWriteSingle(segments[4].Segment, segments[4].Port, displayMapBits[10]);
	}
}
void showTimer(double val)
{
	String s = String(val, 3); //second parameter is number of digits after the decimal
	//Pad string left to 5 characters
	s = stringPADL(s, 5);
	s.replace(".", ""); //more the decimal

	//Clear segments
	clearAllSegments();
	int sL = s.length();
	for (int i = 0; i < sL; i++)
	{
		//Don't show spaces/blanks
		if (s[i] != ' '){
			if (segments[i].Port == GPIOA){
				//Dont need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, displayMap[(int)(s[i] - 48)]);
			}
			else{
				//Need to reverse bits
				expanderWriteSingle(segments[i].Segment, segments[i].Port, reverse(displayMap[(int)(s[i] - 48)]));
			}
		}
	}


}


void loop()
{
	//testDisplay();

	//C# test client located here:
	//C:\Users\Russ\Documents\Visual Studio 2013\Projects\TelnetToEsp\TelnetToEsp
	if (server.hasClient()){
		//Serial.println("Got Client Connection");
		//if (wc.connected()){ wc.stop(); }

		wc = server.available();
		//delay(100);
		if (wc.connected()){
			//Serial.println("Available");
			while (!wc.available()){ yield(); }
			String input = wc.readStringUntil('\t');
			wc.stop(); //release the connection

			//Serial.println(input);
			//Input winner example: "w:speed,et,rt" Sample: "000.00","00.000","00.000"
			//Input looser example: "l:speed,et,rt" Sample: "000.00","00.000","00.000"

			String pp;
			switch (input[0])
			{
			case 'w': //Winner info
				pp = input.substring(input.indexOf(':') + 1);
				displayQ[0].mhp = pp.toFloat();

				pp = pp.substring(pp.indexOf(' ') + 1);
				displayQ[0].et = pp.toFloat();

				pp = pp.substring(pp.indexOf(' ') + 1);
				displayQ[0].rt = pp.toFloat();

				displayQ[0].showArrow = true;
				displayQ[0].ready = true;
				displayQ[0].showMPH = true;
				break;
			case 'l': //Looser info

				pp = input.substring(input.indexOf(':') + 1);
				displayQ[1].mhp = pp.toFloat();

				pp = pp.substring(pp.indexOf(' ') + 1);
				displayQ[1].et = pp.toFloat();

				pp = pp.substring(pp.indexOf(' ') + 1);
				displayQ[1].rt = pp.toFloat();

				displayQ[1].showArrow = true;
				displayQ[1].ready = true;
				displayQ[1].showMPH = true;
				break;
			default:
				break;
			}
		}
	}
	else{ //use this loop to show info
		if (millis() - lastDigitSentTime > 3000){

			for (int i = 0; i < 2; i++){
				if (displayQ[i].ready == true){
					if (displayQ[i].showArrow == true){
						displayQ[i].showArrow = false;
						Serial.print("Arrow: ");
						showArrow(i);
						break;
					}
					if (displayQ[i].showMPH == true){
						displayQ[i].showMPH = false;
						displayQ[i].showET = true;
						showMPH(displayQ[i].mhp);
						break;
					}
					else if (displayQ[i].showET == true){
						Serial.print("ET: ");
						displayQ[i].showET = false;
						displayQ[i].showRT = true;
						showET_OR_RT(displayQ[i].et);
						break;
					}
					else if (displayQ[i].showRT == true){
						Serial.print("RT: ");
						displayQ[i].showRT = false;
						showET_OR_RT(displayQ[i].rt);
						break;

					}
					else{
						displayQ[i].ready = false;
						clearAllSegments();
					}
				}

			}
			lastDigitSentTime = millis();
		}

	}
}
void testDisplay(){
	for (int i = 0; i < 4; i++){
		showArrow(0);
		delay(250);
		clearAllSegments();
		delay(250);
	}

	showET_OR_RT(3.311);
	delay(2000);


	for (int i = 0; i < 4; i++){
		showArrow(1);
		delay(250);
		clearAllSegments();
		delay(250);
	}

	showET_OR_RT(3.971);
	delay(2000);
}

/*
Server Client communications
Commands that can be sent to this module
1) C[lear]:millis, clear the board for the specified time in millis
2) D[isplay]:text, display the specified info for sepcified time in millis
3) M[ode]:mode, do what the mode specifys
4) R, start race started. dot chase across display
5) I, idle, show temp and humidity or other pre programmed messages
6) V, verify board, show each pattern on each segment 1 at a time
7) T, show time HH:MM
8) S, sleep make board blank until told what to do

The client tells this module exactly what to display and for how long in millis.

*/
